"""
Sky statistics computation class for `~skymatch.skymatch` and
`~skymatch.skymatch._weighted_sky`.

:Authors: Mihai Cara

:License: :doc:`LICENSE`

"""
# THIRD PARTY
from stsci.imagestats import ImageStats
from copy import deepcopy

__all__ = ['SkyStats']
__taskname__ = 'skystatistics'


class SkyStats(object):
    """
    This is a superclass build on top of
    :py:class:`stsci.imagestats.ImageStats`. Compared to ``ImageStats``,
    ``SkyStats`` has "persistent settings" in the sense
    that object's parameters need to be set once and these settings
    will be applied to all subsequent computations on different data.

    """
    def __init__(self, skystat='mean', **kwargs):
        """ Initializes the SkyStats object.

        Parameters
        -----------
        skystat: str
            Sets the statistics that will be returned by the
            `~SkyStats.calc_sky`. The following statistics are supported:
            ``'mean'``, ``'mode'``, `'midpt'`, and ``'median'``.
            First three statistics have the same meaning as in
            `stsdas.toolbox.imgtools.gstatistics <http://stsdas.stsci.edu/
            cgi-bin/gethelp.cgi?gstatistics>`_
            while ``skystat='median'`` will compute the median of the
            distribution.

        kwargs: dict
            A dictionary of optional arguments to be passed to ``ImageStats``.

        """
        self.npix = None
        self.skyval = None

        self._fields = ','.join(['npix', skystat])

        self._kwargs = deepcopy(kwargs)
        if 'fields' in self._kwargs:
            del self._kwargs['fields']

        if 'image' in self._kwargs:
            del self._kwargs['image']

        self._skystat = {
            'mean': self._extract_mean,
            'mode': self._extract_mode,
            'median': self._extract_median,
            'midpt': self._extract_midpt
        }[skystat]

    def _extract_mean(self, imstat):
        return imstat.mean

    def _extract_median(self, imstat):
        return imstat.median

    def _extract_mode(self, imstat):
        return imstat.mode

    def _extract_midpt(self, imstat):
        return imstat.midpt

    def calc_sky(self, data):
        """ Computes statistics on data.

        Parameters
        -----------
        data: numpy.ndarray
            A numpy array of values for which the statistics needs to be
            computed.

        Returns
        --------
        statistics: tuple
            A tuple of two values: (``skyvalue``, ``npix``), where ``skyvalue``
            is the statistics specified by the `skystat` parameter during
            the initialization of the ``SkyStats`` object and ``npix`` is the
            number of pixels used in comuting the statistics reported in
            ``skyvalue``.

        """
        imstat = ImageStats(image=data, fields=self._fields, **(self._kwargs))
        self.skyval = self._skystat(imstat)
        self.npix = imstat.npix
        return (self.skyval, self.npix)



"""
This module provides information about supported HST instruments for use
by :py:mod:`stsci.skypac` module.

:Authors: Mihai Cara

:License: :doc:`LICENSE`

"""
supported_telescopes = ['HST']

counts_only_instruments = ['WFPC', 'WFPC2', 'ACS', 'STIS']  # 'HRS'
mixed_units_instruments = ['NICMOS', 'WFC3']
rates_only_instruments = ['FOC', 'COS']

supported_instruments = (
    counts_only_instruments + mixed_units_instruments + rates_only_instruments
)

photcorr_kwd = {
    'FOC': ['WAVCORR', 'COMPLETE'],
    'WFPC': ['DOPHOTOM', 'DONE'],
    'WFPC2': ['DOPHOTOM', 'COMPLETE'],
    'NICMOS': ['PHOTDONE', 'PERFORMED'],
    'STIS': ['PHOTCORR', 'COMPLETE'],
    'ACS': ['PHOTCORR', 'COMPLETE'],
    'WFC3': ['PHOTCORR', 'COMPLETE'],
    'COS': ['PHOTCORR', 'COMPLETE'],
}



"""
A module that provides functions for computing Pixel Area Maps (PAM) based
on polynomial distortion model contained in a FITS WCS. Tabular distortions
``NPOL`` and ``DET2IM`` used to describe ``HST/ACS`` distortions are ignored.

:Authors: Mihai Cara

:License: :doc:`LICENSE`

"""
import numbers
import numpy as np
from astropy.io import fits
from astropy import wcs as fitswcs
import stwcs


__all__ = ['pam_from_file', 'pam_from_wcs']


def _is_int(n):
    return (
        (isinstance(n, numbers.Integral) and not isinstance(n, bool)) or
        (isinstance(n, np.generic) and np.issubdtype(n, np.integer))
    )


def _compute_pam_sd(wcs, shape=None, blc=(1, 1), idcscale=1.0, cdscale=1.0):
    """
    Computes Pixel Area Map (PAM) using the distortion model defined in WCS
    and described through Simple Image Polynomials (SIP) by computing
    the Jacobian of the distortion model.

    This function computes the Jacobian of the distortion model using
    *symbolic differentiation* of Simple Image Polynomials.

    Parameters
    ----------

    wcs: astropy.wcs.WCS, stwcs.wcsutil.HSTWCS
        A ``WCS`` object containing the distortion model.

    shape: tuple of int, None, optional
        A tuple of two integers (ny, nx) indicating the size of the PAM image
        to be generated. When the default value is used (`None`), the size
        of the returned PAM array will be determined from ``wcs.array_shape``
        attribute of the supplied ``WCS`` object.

    blc: tuple of int or float, optional
        A tuple indicating the coordinates of the bottom-left pixel of the
        PAM array to be computed. These coordinates should be given
        in the image coordinate system defined by the input ``WCS`` (in which,
        for example, ``WCS.crpix`` is defined). The first element specifies
        the column (``"x"``-coordinate) and the second element specifies
        the row (``"y"``-coordinate).

    idcscale: float, optional
        A positive number indicating the pixel scale used in the
        "Instrument Distortion Correction" for HST instruments. For
        non-HST instruments this parameter may be set to be equal
        to ``cdscale``.

    cdscale: float, optional
        A positive number indicating the pixel scale as computed from the
        CD matrix. HST instruments CD matrix includes linear distortion
        terms.

    Returns
    -------
    PAM: numpy.ndarray
        Pixel area map.

    """
    if shape is None:
        shape = wcs.array_shape

    # rescale factor:
    rf = (cdscale / idcscale)**2

    # distortion does not exist or is linear:
    if wcs.sip is None or wcs.sip.a_order < 1 or wcs.sip.b_order < 1 or \
       (wcs.sip.a_order == 1 and wcs.sip.b_order == 1):
        return rf * np.ones(shape, dtype=np.float64)

    # prepare coordinates:
    x = np.arange(shape[1], dtype=float) - wcs.sip.crpix[0] + float(blc[0])
    y = np.arange(shape[0], dtype=float) - wcs.sip.crpix[1] + float(blc[1])

    ar = np.arange(wcs.sip.a_order + 1)
    br = np.arange(wcs.sip.b_order + 1)

    ones_a = np.ones(wcs.sip.a_order + 1)
    ones_b = np.ones(wcs.sip.b_order + 1)

    # "coordinate vectors" (e.g., (1, x, x**2, x**3, ...)) used in
    # distortion bilinear forms:
    ax = np.outer(x, ones_a)**ar
    ay = np.outer(y, ones_a)**ar
    bx = np.outer(x, ones_b)**br
    by = np.outer(y, ones_b)**br

    # derivatives of the "coordinate vectors" with regard to x & y:
    adx = np.roll(ax, 1, 1) * ar
    ady = np.roll(ay, 1, 1) * ar
    bdx = np.roll(bx, 1, 1) * br
    bdy = np.roll(by, 1, 1) * br

    # derivatives of the binomial forms:
    A = wcs.sip.a.T
    B = wcs.sip.b.T
    dadx = 1.0 + np.tensordot(ay.T, np.tensordot(A, adx, (1, 1)), (0, 0))
    dady = np.tensordot(ady.T, np.tensordot(A, ax, (1, 1)), (0, 0))
    dbdx = np.tensordot(by.T, np.tensordot(B, bdx, (1, 1)), (0, 0))
    dbdy = 1.0 + np.tensordot(bdy.T, np.tensordot(B, bx, (1, 1)), (0, 0))

    # compute rescaled Jacobian
    jacobian = rf * np.abs(dadx * dbdy - dady * dbdx)

    return jacobian


def pam_from_file(image, ext, output_pam, ignore_vacorr=False,
                  normalize_at=None):
    r"""
    Generate a **P**\ ixel **A**\ rea **M**\ ap (PAM) file from the ``FITS``
    ``WCS`` contained in an image extension of a calibrated ``HST`` image
    file specified by ``image``.

    .. note::
       PAM computation is performed using the distortion model defined in the
       ``WCS`` and described through Simple Image Polynomials (SIP).
       Non-polynomial distortions are ignored!

    Parameters
    ----------

    image: str
        File name of a ``FITS`` image that will provide a ``FITS`` ``WCS``
        (`stwcs.wcsutils.HSTWCS` or `astropy.wcs.WCS`).

    ext: int, str, tuple of (str, int)
        Extension specification. May be an integer extension number,
        a string extension name, or a tuple of extension name *and*
        extension version.

    output_pam: str
        Output file name to which PAM will be written.

        .. warning::
            If the output file already exists, it will be overwritten
            without warnings.

    ignore_vacorr: bool, optional
        When set to `True`, ``PAM`` will be generated *as if* vellocity
        aberration has not applied to the ``WCS``.

        .. warning::
           This function does not know whether velocity aberration (VA)
           correction has been applied to the ``WCS`` or not. It is user's
           responsibility to check the appropriateness of settung this
           parameter to `True`. Setting ``ignore_vacorr`` to `True` when
           ``WCS`` was not VA-corrected will result in larger errors in
           computed ``PAM``. **Default value is highly recommended!**

    normalize_at: tuple of int, optional
        Indicates whether to normalize computed ``PAM`` to 1 at the provided
        zero-based pixel position. By default, PAM is computed relative to
        (or, in units of) the ``idcscale`` (for ``HST`` instruments) value
        when present or to the pixel scale at ``CRPIX`` when the ``wcs``
        object does not have an ``idcscale`` property. Default setting
        should produce results analogous to the drizzle/blot method.

        .. note::
           ``HST/WFC3`` PAM historically are normalized to 1 at specific pixel
           positions. See http://www.stsci.edu/hst/wfc3/pam/pixel_area_maps for
           further details.

    """
    with fits.open(image, mode='readonly') as h:
        data_shape = h[ext].data.shape
        try:
            wcs = stwcs.wcsutil.HSTWCS(h, ext)
        except Exception:
            wcs = fitswcs.WCS(h[ext].header, h)

    pam = pam_from_wcs(wcs, shape=data_shape, ignore_vacorr=ignore_vacorr,
                       normalize_at=normalize_at)

    fits.PrimaryHDU(pam).writeto(output_pam, overwrite=True)


def pam_from_wcs(wcs, shape=None, ignore_vacorr=False,
                 normalize_at=None):
    r"""
    Generate a **P**\ ixel **A**\ rea **M**\ ap (PAM) file from a ``FITS``
    ``WCS``.

    .. note::
       PAM computation is performed using the distortion model defined in the
       ``WCS`` and described through Simple Image Polynomials (SIP).
       Non-polynomial distortions are ignored!

    Parameters
    ----------

    wcs: stwcs.wcsutils.HSTWCS, astropy.wcs.WCS
        An `~astropy.wcs.WCS` object to be used for generating PAM file.

    shape: tuple of two int, None, optional
        Shape of the output image ``(ny, nx)``. If se to default `None`, this
        function will try to deduce the shape of the output image from the
        value of ``array_shape`` attribute of the input ``wcs`` object.

    ignore_vacorr: bool, optional
        When set to `True`, ``PAM`` will be generated *as if* vellocity
        aberration has not applied to the ``WCS``.

        .. warning::
           This function does not know whether velocity aberration (VA)
           correction has been applied to the ``WCS`` or not. It is user's
           responsibility to check the appropriateness of settung this
           parameter to `True`. Setting ``ignore_vacorr`` to `True` when
           ``WCS`` was not VA-corrected will result in larger errors in
           computed ``PAM``. **Default value is highly recommended!**

    normalize_at: tuple of int, optional
        Indicates whether to normalize computed ``PAM`` to 1 at the provided
        zero-based pixel position. By default, PAM is computed relative to
        (or, in units of) the ``idcscale`` (for ``HST`` instruments) value
        when present or to the pixel scale at ``CRPIX`` when the ``wcs``
        object does not have an ``idcscale`` property. Default setting
        should produce results analogous to the drizzle/blot method.

        .. note::
           ``HST/WFC3`` PAM historically are normalized to 1 at specific pixel
           positions. See http://www.stsci.edu/hst/wfc3/pam/pixel_area_maps for
           further details.

    Returns
    -------

    pam: numpy.ndarray
        A 2D `numpy.ndarray` containing PAM.

    Raises
    ------

    ValueError
        When both ``shape`` and ``wcs.array_shape`` attribute are `None`.

    """
    if shape is None:
        if wcs.array_shape is not None:
            shape = wcs.array_shape
        else:
            raise ValueError("Both 'shape' and 'wcs.array_shape' are None. "
                             "Cannot infer output image shape.")

    cdscale = np.sqrt(fitswcs.utils.proj_plane_pixel_area(wcs)) * 3600.0

    # Ignoring VAFACTOR helps get PAMs that are in better
    # agreement with the ones produced by R. Hook in 2004.
    if ignore_vacorr and hasattr(wcs, 'vafactor'):
        cdscale /= wcs.vafactor

    idcscale = wcs.idcscale if hasattr(wcs, 'idcscale') else cdscale
    pam = _compute_pam_sd(wcs=wcs, shape=shape, idcscale=idcscale,
                          cdscale=cdscale)

    if normalize_at is not None:
        if hasattr(normalize_at, '__iter__'):
            if len(normalize_at) != 2:
                raise ValueError("'normalize_at' must contain exactly two "
                                 "coordinates.")
            if not all([_is_int(coord) for coord in normalize_at]):
                raise TypeError("Each pixel coordinate must be an integer.")
            if normalize_at[0] >= shape[1] or normalize_at[1] >= shape[0] or \
               normalize_at[0] < 0 or normalize_at[1] < 0:
                raise ValueError(
                    "Pixel coordinates specified by 'normalize_at' must be "
                    "within output image borders.")
        else:
            raise TypeError(
                "When specified, 'normalize_at' must be an iterable."
            )

        pam /= pam[normalize_at[1], normalize_at[0]]

    return pam



"""skymatch"""
import os
from pkg_resources import get_distribution, DistributionNotFound

try:
    __version__ = get_distribution(__name__).version
except DistributionNotFound:
    # package is not installed
    __version__ = 'UNKNOWN'

# from .version import version as __version__

__author__ = 'Mihai Cara'
__docformat__ = 'restructuredtext en'
__taskname__ = 'skymatch'


from . import parseat  # noqa: F401
from . import utils  # noqa: F401
from . import pamutils  # noqa: F401
from . import region  # noqa: F401
from . import skystatistics  # noqa: F401
from . import skyline  # noqa: F401
from . import skymatch  # noqa: F401

from stsci.tools import teal
teal.print_tasknames(__name__, os.path.dirname(__file__))


def help():
    msg = """The SkyPac package contains the following tasks that allow users
    perform sky level matching on user images.

    skypac:
        skymatch - primary task for performing sky level matching on
                   user images
    """
    print(msg)



#!/usr/bin/env python
import os
import pkgutil
import sys
from setuptools import setup, find_packages
from subprocess import check_call, CalledProcessError

try:
    from distutils.config import ConfigParser
except ImportError:
    from configparser import ConfigParser

conf = ConfigParser()
conf.read(['setup.cfg'])

# Get some config values
metadata = dict(conf.items('metadata'))
PACKAGENAME = metadata.get('package_name', 'stsci.skypac')
DESCRIPTION = metadata.get('description', 'Sky matching on image mosaic')
LONG_DESCRIPTION = metadata.get('long_description', 'README.md')
LONG_DESCRIPTION_CONTENT_TYPE = metadata.get('long_description_content_type',
                                             'text/markdown')
AUTHOR = metadata.get('author', 'Mihai Cara, Warren Hack, Pey Lian Lim')
AUTHOR_EMAIL = metadata.get('author_email', 'help@stsci.edu')
URL = metadata.get('url', 'https://github.com/spacetelescope/stsci.skypac')
LICENSE = metadata.get('license', 'BSD-3-Clause')

# load long description
this_dir = os.path.abspath(os.path.dirname(__file__))
with open(os.path.join(this_dir, LONG_DESCRIPTION), encoding='utf-8') as f:
    long_description = f.read()

PACKAGE_DATA = {
    '': [
        'README.rst',
        'LICENSE.txt',
        'CHANGELOG.rst',
        '*.fits',
        '*.txt',
        '*.inc',
        '*.cfg',
        '*.csv',
        '*.yaml',
        '*.json'
        ],

    PACKAGENAME: [
        '*.help',
        'pars/*',
    ],
}

INSTALL_REQUIRES = [
    'astropy>=3.1',
    'numpy',
    'spherical_geometry>=1.2.2',
    'stsci.imagestats',
    'stsci.tools',
    'stwcs',
]

SETUP_REQUIRES = [
    'setuptools_scm',
    'pytest-runner',
]

TESTS_REQUIRE = [
    'pytest',
    'pytest-cov',
    'pytest-doctestplus',
    'codecov',
]

DOCS_REQUIRE = [
    'numpydoc',
    'graphviz',
    'sphinx<=1.8.5',
    'sphinx_rtd_theme',
    'stsci_rtd_theme',
    'sphinx_automodapi',
]

setup(
    name=PACKAGENAME,
    author=AUTHOR,
    author_email=AUTHOR_EMAIL,
    description=DESCRIPTION,
    long_description=long_description,
    long_description_content_type=LONG_DESCRIPTION_CONTENT_TYPE,
    url=URL,
    classifiers=[
        'Intended Audience :: Science/Research',
        'License :: OSI Approved :: BSD License',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Topic :: Scientific/Engineering :: Astronomy',
        'Topic :: Software Development :: Libraries :: Python Modules',
        'Development Status :: 5 - Production/Stable',
    ],
    use_scm_version={'write_to': 'stsci/skypac/version.py'},
    setup_requires=SETUP_REQUIRES,
    python_requires='>=3.5',
    install_requires=INSTALL_REQUIRES,
    tests_require=TESTS_REQUIRE,
    packages=find_packages(),
    package_data=PACKAGE_DATA,
    ext_modules=[],
    extras_require={
        'docs': DOCS_REQUIRE,
        'test': TESTS_REQUIRE,
        'all': DOCS_REQUIRE + TESTS_REQUIRE,
    },
    project_urls={
        'Bug Reports': 'https://github.com/spacetelescope/stsci.skypac/issues/',
        'Source': 'https://github.com/spacetelescope/stsci.skypac/',
        'Help': 'https://hsthelp.stsci.edu/',
    },
)



"""
Polygon filling algorithm.

:Authors: Nadezhda Dencheva, Mihai Cara

:License: :doc:`LICENSE`

"""
# Original author: Nadezhda Dencheva
#
# modifications by Mihai Cara: removed functionality not needed for the
# skymatch algorithm and modified the code to be able to work with polygons
# that have vertices with negative coordinates. Polygon vertices are now
# *internally* (to region.py) rounded to integers so that Polygon will not
# crash when input vertices are floats. Fixed a bug in _construct_ordered_GET
# that was causing varying polygon filling for different ordering of the
# vertices. Finally, modified the algorithm to fill the right-most pixels
# as well as top-most row of the polygon.
#
# NOTE: Algorithm description can be found, e.g., here:
#    http://www.cs.rit.edu/~icss571/filling/how_to.html
#    http://www.cs.uic.edu/~jbell/CourseNotes/ComputerGraphics/PolygonFilling.html
#
from collections import OrderedDict
import numpy as np

__all__ = ['Region', 'Edge', 'Polygon']
__taskname__ = 'region'


class ValidationError(Exception):
    def __init__(self, message):
        self._message = message

    def __str__(self):
        return self._message


class Region(object):
    """
    Base class for regions.

    Parameters
    -------------
    rid: int or string
        region ID
    coordinate_system: astropy.wcs.CoordinateSystem instance or a string
        in the context of WCS this would be an instance of wcs.CoordinateSysem
    """
    def __init__(self, rid, coordinate_system):
        self._coordinate_system = coordinate_system
        self._rid = rid

    def __contains__(self, x, y):
        """
        Determines if a pixel is within a region.

        Parameters
        ----------
        x, y: float
            x , y values of a pixel

        Returns
        -------
        True or False

        Subclasses must define this method.
        """
        raise NotImplementedError("__contains__")

    def scan(self, mask):
        """
        Sets mask values to region id for all pixels within the region.
        Subclasses must define this method.

        Parameters
        ----------
        mask: ndarray
            a byte array with the shape of the observation to be used as a mask

        Returns
        -------
        mask: array where the value of the elements is the region ID or 0 (for
            pixels which are not included in any region).

        """
        raise NotImplementedError("scan")


class Polygon(Region):
    """
    Represents a 2D polygon region with multiple vertices

    Parameters
    ----------
    rid: string
         polygon id

    vertices: list of (x,y) tuples or lists
         The list is ordered in such a way that when traversed in a
         counterclockwise direction, the enclosed area is the polygon.
         The last vertex must coincide with the first vertex, minimum
         4 vertices are needed to define a triangle.

    coord_system: string
        coordinate system

    """
    def __init__(self, rid, vertices, coord_system="Cartesian"):
        assert len(vertices) >= 4, ("Expected vertices to be "
                                    "a list of minimum 4 tuples (x,y)")
        super(Polygon, self).__init__(rid, coord_system)

        # self._shiftx & self._shifty are introduced to shift the bottom-left
        # corner of the polygon's bounding box to (0,0) as a (hopefully
        # temporary) workaround to a limitation of the original code that the
        # polygon must be completely contained in the image. It seems that the
        # code works fine if we make sure that the bottom-left corner of the
        # polygon's bounding box has non-negative coordinates.
        self._shiftx = 0
        self._shifty = 0
        for vertex in vertices:
            x, y = vertex
            if x < self._shiftx:
                self._shiftx = x
            if y < self._shifty:
                self._shifty = y
        v = [(i - self._shiftx, j - self._shifty) for i, j in vertices]

        # convert to integer coordinates:
        self._vertices = np.asarray(list(map(_round_vertex, v)))
        self._shiftx = int(round(self._shiftx))
        self._shifty = int(round(self._shifty))

        self._bbox = self._get_bounding_box()
        self._scan_line_range = list(
            range(self._bbox[1], self._bbox[3] + self._bbox[1] + 1)
        )
        # constructs a Global Edge Table (GET) in bbox coordinates
        self._GET = self._construct_ordered_GET()

    def _get_bounding_box(self):
        x = self._vertices[:, 0].min()
        y = self._vertices[:, 1].min()
        w = self._vertices[:, 0].max() - x
        h = self._vertices[:, 1].max() - y
        return (x, y, w, h)

    def _construct_ordered_GET(self):
        """
        Construct a Global Edge Table (GET)

        The GET is an OrderedDict. Keys are scan  line numbers,
        ordered from ``bbox.ymin`` to ``bbox.ymax``, where ``bbox`` is the
        bounding box of the polygon.
        Values are lists of edges for which ``edge.ymin==scan_line_number``.

        Returns
        -------
        GET: OrderedDict
            {scan_line: [edge1, edge2]}

        """
        # edges is a list of Edge objects which define a polygon
        # with these vertices
        edges = self.get_edges()
        GET = OrderedDict.fromkeys(self._scan_line_range)
        ymin = np.asarray([e._ymin for e in edges])
        for i in self._scan_line_range:
            ymin_ind = (ymin == i).nonzero()[0]
            yminindlen, = ymin_ind.shape

            # if ymin_ind.any(): # original
            # mcara - a hack for incomplete filling .any() fails
            # if 0 is in ymin_ind
            if yminindlen:
                GET[i] = [edges[ymin_ind[0]]]
                for j in ymin_ind[1:]:
                    GET[i].append(edges[j])
        return GET

    def get_edges(self):
        """
        Create a list of Edge objects from vertices
        """
        edges = []
        for i in range(1, len(self._vertices)):
            name = 'E' + str(i - 1)
            edges.append(
                Edge(name=name, start=self._vertices[i - 1],
                     stop=self._vertices[i])
            )
        return edges

    def scan(self, data):
        """
        This is the main function which scans the polygon and creates the mask

        Parameters
        ----------
        data: array
            the mask array
            it has all zeros initially, elements within a region are set to
            the region's ID

        Notes
        -----
        Algorithm summary:

        - Set the Global Edge Table (GET)
        - Set y to be the smallest y coordinate that has an entry in GET
        - Initialize the Active Edge Table (AET) to be empty
        - For each scan line:

          1. Add edges from GET to AET for which ymin==y
          2. Remove edges from AET fro which ymax==y
          3. Compute the intersection of the current scan line with all
             edges in the AET
          4. Sort on X of intersection point
          5. Set elements between pairs of X in the AET to the Edge's ID

        """
        # # TODO:
        # # 1. This algorithm does not mark pixels in the top row and left
        # #    most column. Pad the initial pixel description on top and left
        # #    with 1 px to prevent this.
        # # 2. Currently it uses intersection of the scan line with edges.
        # #    If this is too slow it should use the 1/m increment (replace 3
        # #    above) (or the increment should be removed from the GET entry).

        # see comments in the __init__ function for the reason of introducing
        # polygon shifts (self._shiftx & self._shifty). Here we need to shift
        # it back.

        (ny, nx) = data.shape

        y = np.min(list(self._GET.keys()))

        AET = []
        scline = self._scan_line_range[-1]
        while y <= scline:
            if y < scline:
                AET = self.update_AET(y, AET)

            if self._bbox[2] <= 0:
                y += 1
                continue

            scan_line = Edge('scan_line', start=[self._bbox[0], y],
                             stop=[self._bbox[0] + self._bbox[2], y])

            x = [int(np.ceil(e.compute_AET_entry(scan_line)[1]))
                 for e in AET if e is not None]

            xnew = np.sort(x)

            if y + self._shifty < 0 or y + self._shifty >= ny:
                y += 1
                continue

            for i, j in zip(xnew[::2], xnew[1::2]):
                xstart = i + self._shiftx if (i + self._shiftx) >= 0 else 0
                xend = j + self._shiftx if (j + self._shiftx) < nx else nx - 1
                data[y + self._shifty][xstart:xend + 1] = self._rid

            y += 1

        return data

    def update_AET(self, y, AET):
        """
        Update the Active Edge Table (AET)

        Add edges from GET to AET for which ymin of the edge is
        equal to the y of the scan line.
        Remove edges from AET for which ymax of the edge is
        equal to y of the scan line.

        """
        edge_cont = self._GET[y]
        if edge_cont is not None:
            for edge in edge_cont:
                if edge._start[1] != edge._stop[1] and edge._ymin == y:
                    AET.append(edge)
        for edge in AET[::-1]:
            if edge is not None:
                if edge._ymax == y:
                    AET.remove(edge)
        return AET

    def __contains__(self, px):
        """even-odd algorithm or smth else better sould be used"""
        # minx = self._vertices[:,0].min()
        # maxx = self._vertices[:,0].max()
        # miny = self._vertices[:,1].min()
        # maxy = self._vertices[:,1].max()
        return (
            px[0] >= self._bbox[0] and
            px[0] <= self._bbox[0] + self._bbox[2] and
            px[1] >= self._bbox[1] and
            px[1] <= self._bbox[1] + self._bbox[3]
        )


class Edge(object):
    """
    Edge representation

    An edge has a "start" and "stop" ``(x, y)`` vertices and an entry in the
    GET table of a polygon. The GET entry is a list of these values:

    ``[ymax, x_at_ymin, delta_x/delta_y]``

    """
    def __init__(self, name=None, start=None, stop=None, next=None):
        self._start = None
        if start is not None:
            self._start = np.asarray(start)
        self._name = name
        self._stop = stop
        if stop is not None:
            self._stop = np.asarray(stop)
        self._next = next

        if self._stop is not None and self._start is not None:
            if self._start[1] < self._stop[1]:
                self._ymin = self._start[1]
                self._yminx = self._start[0]
            else:
                self._ymin = self._stop[1]
                self._yminx = self._stop[0]

            self._ymax = max(self._start[1], self._stop[1])
            self._xmin = min(self._start[0], self._stop[0])
            self._xmax = max(self._start[0], self._stop[1])

        else:
            self._ymin = None
            self._yminx = None
            self._ymax = None
            self._xmin = None
            self._xmax = None

        self.GET_entry = self.compute_GET_entry()

    @property
    def ymin(self):
        return self._ymin

    @property
    def start(self):
        return self._start

    @property
    def stop(self):
        return self._stop

    @property
    def ymax(self):
        return self._ymax

    def compute_GET_entry(self):
        """
        Compute the entry in the Global Edge Table

        ``[ymax, x@ymin, 1/m]``

        """
        if self._start is None:
            return None

        earr = np.asarray([self._start, self._stop])
        if np.diff(earr[:, 1]).item() == 0:
            return None

        entry = [
            self._ymax,
            self._yminx,
            (np.diff(earr[:, 0]) / np.diff(earr[:, 1])).item(),
            None
        ]
        return entry

    def compute_AET_entry(self, edge):
        """
        Compute the entry for an edge in the current Active Edge Table

        [ymax, x_intersect, 1/m]
        note: currently 1/m is not used
        """
        x = self.intersection(edge)[0]
        return [self._ymax, x, self.GET_entry[2]]

    def __repr__(self):
        fmt = ""
        if self._name is not None:
            fmt += self._name
            ne = self.next_edge
            while ne is not None:
                fmt += "-->"
                fmt += ne._name
                ne = ne.next_edge
        return fmt

    @property
    def next_edge(self):
        return self._next

    @next_edge.setter
    def next_edge(self, edge):
        if self._name is None:
            self._name = edge._name
            self._stop = edge._stop
            self._start = edge._start
            self._next = edge.next_edge
        else:
            self._next = edge

    def intersection(self, edge):
        u = self._stop - self._start
        v = edge._stop - edge._start
        w = self._start - edge._start
        D = np.cross(u, v)
        if np.allclose(np.cross(u, v), 0, rtol=0,
                       atol=1e2 * np.finfo(float).eps):
            return np.array(self._start)

        return np.cross(v, w) / D * u + self._start

    def is_parallel(self, edge):
        u = self._stop - self._start
        v = edge._stop - edge._start
        return np.allclose(np.cross(u, v), 0, rtol=0,
                           atol=1e2 * np.finfo(float).eps)


def _round_vertex(v):
    x, y = v
    return (int(round(x)), int(round(y)))
